"""Dummy models to be used in test cases of the ``privacy`` app."""
from django.core.urlresolvers import reverse
from django.db import models


class DummyModel(models.Model):
    """Dummy to be used in test cases of the ``privacy`` app."""
    name = models.CharField(max_length=256, blank=True)
    privacy = models.ForeignKey(
        'privacy.PrivacySetting', blank=True, null=True)

    def __unicode__(self):
        return self.name


class DummyProfileModel(models.Model):
    """Dummy profile to be used in test cases of the ``privacy`` app."""
    user = models.ForeignKey('auth.User', related_name='dummy_profiles')
    friends = models.ManyToManyField('auth.User', blank=True, null=True,
                                     related_name='friended_dummy_profiles')

    def get_absolute_url(self):
        return reverse('dummy_profile_update', kwargs={'pk': self.pk})



"""Forms for the ``privacy`` app."""
from django.contrib.contenttypes.models import ContentType

from .models import PrivacyLevel, PrivacySetting


class PrivacyFormMixin(object):
    """Form mixin for forms with privacy level fields."""
    def save(self, *args, **kwargs):
        instance = super(PrivacyFormMixin, self).save(*args, **kwargs)
        for key in self.data.keys():
            if not key.startswith('privacy_'):
                continue
            field_name = key.replace('privacy_', '')
            if field_name == 'for_instance':
                field_name = ''
            ctype = ContentType.objects.get_for_model(self.instance)
            kwargs = {
                'content_type': ctype,
                'object_id': self.instance.pk,
                'field_name': field_name,
            }
            level = PrivacyLevel.objects.get(clearance_level=self.data[key])
            try:
                privacy_setting = PrivacySetting.objects.get(**kwargs)
            except PrivacySetting.DoesNotExist:
                kwargs.update({'level': level})
                privacy_setting = PrivacySetting.objects.create(**kwargs)
            else:
                privacy_setting.level = level
                privacy_setting.save()
        return instance



"""Tests for the models of the ``privacy`` app."""
from django.test import TestCase

from .factories import (
    PrivacyLevelFactory,
    PrivacySettingFactory,
)
from ..models import filter_privacy_level
from test_app.factories import DummyModelFactory
from test_app.models import DummyModel


class PrivacyLevelTestCase(TestCase):
    """Tests for the ``PrivacyLevel`` model."""
    longMessage = True

    def setUp(self):
        self.privacy_level = PrivacyLevelFactory()

    def test_model(self):
        self.assertTrue(self.privacy_level.pk, msg=(
            'Should be able to instantiate and save the object.'))


class PrivacySettingTestCase(TestCase):
    """Tests for the ``PrivacySetting`` model."""
    longMessage = True

    def setUp(self):
        self.privacy_setting = PrivacySettingFactory()

    def test_model(self):
        self.assertTrue(self.privacy_setting.pk, msg=(
            'Should be able to instantiate and save the object.'))


class FilterPrivacyLevelTestCase(TestCase):
    """Tests for the ``filter_privacy_level`` helper function."""
    longMessage = True

    def setUp(self):
        self.dummy_1 = DummyModelFactory()
        self.dummy_2 = DummyModelFactory()
        self.dummy_3 = DummyModelFactory()

    def test_model(self):
        qs = []
        self.assertFalse(filter_privacy_level(qs, 4), msg=(
            'Should return the empty queryset, instead of raising an'
            ' exception, if queryset has no objects.'))
        qs = DummyModel.objects.all()
        self.assertEqual(filter_privacy_level(qs, 4).count(), 3, msg=(
            'Should return all objects, if they got no privacy settings.'))
        PrivacySettingFactory(content_object=self.dummy_1,
                              level__clearance_level=1)
        PrivacySettingFactory(content_object=self.dummy_2,
                              level__clearance_level=2)
        PrivacySettingFactory(content_object=self.dummy_3,
                              level__clearance_level=3)
        self.assertEqual(filter_privacy_level(qs, 1).count(), 1, msg=(
            'Should return only objects, which got level 1 or lower.'))
        self.assertEqual(filter_privacy_level(qs, 2).count(), 2, msg=(
            'Should return only objects, which got level 2 or lower.'))
        self.assertEqual(filter_privacy_level(qs, 3).count(), 3, msg=(
            'Should return only objects, which got level 3 or lower.'))
        self.assertEqual(filter_privacy_level(qs, 3, True).count(), 1, msg=(
            'Should return only objects, which got level 3 exactly.'))



"""Forms for the ``test_app`` app."""
from django import forms

from .models import DummyProfileModel
from privacy.forms import PrivacyFormMixin


class DummyProfileModelForm(PrivacyFormMixin, forms.ModelForm):
    class Meta:
        model = DummyProfileModel
        fields = ('user', )



"""Admin for the ``privacy`` app."""
from django.contrib import admin

from hvad.admin import TranslatableAdmin

from .models import PrivacyLevel, PrivacySetting


admin.site.register(PrivacyLevel, TranslatableAdmin)
admin.site.register(PrivacySetting, admin.ModelAdmin)



"""Models for the ``privacy`` app."""
from django.conf import settings
from django.db import models
from django.contrib.contenttypes.models import ContentType
from django.contrib.contenttypes import generic
from django.utils.translation import ugettext_lazy as _

from hvad.models import TranslatableModel, TranslatedFields


def filter_privacy_level(qs, clearance_level, exact=False):
    """
    Function to exclude objects from a queryset, which got a higher clearance
    level than the wanted maximum clearance level.

    :qs: Django queryset.
    :clearance_level: Minimum clearance level.
    :exact: Boolean to check for the exact clearance level.

    """
    if not qs:
        return qs
    c_type = ContentType.objects.get_for_model(qs.model)
    kwargs = {
        'content_type': c_type,
        'object_id__in': qs.values_list('pk'),
        'level__clearance_level{}'.format(
            '' if exact else '__gt'): clearance_level,
    }
    private_objects = PrivacySetting.objects.filter(**kwargs).values_list(
        'object_id')
    if exact:
        return qs.filter(pk__in=private_objects)
    return qs.exclude(pk__in=private_objects)


class PrivacyLevel(TranslatableModel):
    """
    Privacy level, which defines the assessment factor.

    :name: Translatable name of the privacy level.
    :clearance_level: Privacy level (the higher the more secure).

    """
    translations = TranslatedFields(
        name=models.CharField(max_length=255),
    )

    clearance_level = models.IntegerField(
        unique=True,
        verbose_name=_('Clearance level'),
    )

    class Meta:
        ordering = getattr(
            settings, 'PRIVACY_LEVEL_ORDERING', ['clearance_level'])

    def __unicode__(self):
        return self.safe_translation_getter(
            'name', self.translations.all()[0].name)


class PrivacySetting(models.Model):
    """
    Privacy setting, which defines one custom setting for one object.

    :content_object: Related secured object.
    :level: Connection to the PrivacyLevel model.
    :field_name: Optional field_name to limit the setting to a single field.

    """
    content_type = models.ForeignKey(ContentType)
    object_id = models.PositiveIntegerField()
    content_object = generic.GenericForeignKey('content_type', 'object_id')

    level = models.ForeignKey(
        PrivacyLevel,
        verbose_name=_('Privacy level'),
    )

    field_name = models.CharField(
        max_length=50,
        verbose_name=_('Field name'),
        blank=True,
    )

    def __unicode__(self):
        return '{} - {}'.format(self.content_object, self.level)



"""Tests for the template tags of the ``privacy`` app."""
from django.template import Context, Template
from django.test import TestCase

from ..templatetags.privacy_tags import is_access_allowed, get_privacy_setting
from .factories import PrivacyLevelFactory, PrivacySettingFactory
from test_app.factories import DummyProfileModelFactory, DummyModelFactory


class GetPrivacySettingTestCase(TestCase):
    """Tests for the ``get_privacy_setting`` filter tag."""
    longMessage = True

    def setUp(self):
        self.obj = DummyModelFactory()
        self.level_1 = PrivacyLevelFactory(clearance_level=1)
        self.level_2 = PrivacyLevelFactory(clearance_level=2)

    def test_tag(self):
        self.assertIsNone(get_privacy_setting(self.obj), msg=(
            'Should return None, if there\'s no setting set.'))
        PrivacySettingFactory(level=self.level_1, content_object=self.obj)
        self.assertEqual(
            get_privacy_setting(self.obj).level.clearance_level, 1, msg=(
                'Should return the privacy setting.'))


class RenderPrivacyLevelFieldTestCase(TestCase):
    """Tests for the ``render_privacy_level_field`` inclusion tag."""
    longMessage = True

    def setUp(self):
        self.obj = DummyModelFactory()
        self.level_1 = PrivacyLevelFactory(clearance_level=1)
        self.level_2 = PrivacyLevelFactory(clearance_level=2)

    def test_tag(self):
        t = Template(
            '{% load privacy_tags %}{% render_privacy_level_field obj %}')
        c = Context({'obj': self.obj})
        self.assertIn('name="privacy_for_instance"', t.render(c))
        t = Template('{% load privacy_tags %}{% render_privacy_level_field obj'
                     ' "name" %}')
        self.assertIn('name="privacy_name"', t.render(c))


class IsAccessAllowedTestCase(TestCase):
    """Tests for the ``is_access_allowed`` assignment tag."""
    longMessage = True

    def setUp(self):
        self.level_1 = PrivacyLevelFactory(clearance_level=1)
        self.level_2 = PrivacyLevelFactory(clearance_level=2)
        self.level_3 = PrivacyLevelFactory(clearance_level=3)
        self.level_4 = PrivacyLevelFactory(clearance_level=4)
        self.owner = DummyProfileModelFactory()
        self.requester = DummyProfileModelFactory()
        self.obj = DummyModelFactory()

    def test_tag(self):
        # No level
        self.assertTrue(
            is_access_allowed(self.owner, self.requester, self.obj),
            msg=('Access should be allowed, if it\'s an object without privacy'
                 ' setting.'))

        # Level 1
        setting = PrivacySettingFactory(
            level=self.level_1, content_object=self.obj)
        self.assertTrue(
            is_access_allowed(self.owner, self.requester, self.obj),
            msg=('Access should be allowed, if it\'s a public object.'))

        # Level 2
        setting.level = self.level_2
        setting.save()
        self.assertFalse(
            is_access_allowed(self.owner, self.requester, self.obj),
            msg=('Access should not be allowed, if the requester is not at'
                 ' least a friend of a friend.'))
        friend = DummyProfileModelFactory()
        self.owner.friends.add(friend.user)
        self.requester.friends.add(friend.user)
        self.assertTrue(
            is_access_allowed(self.owner, self.requester, self.obj),
            msg=('Access should be allowed, if the requester is at least a'
                 ' friend of a friend.'))

        # Level 3
        setting.level = self.level_3
        setting.save()
        self.assertFalse(
            is_access_allowed(self.owner, self.requester, self.obj),
            msg=('Access should not be allowed, if the requester is not at'
                 ' least a friend.'))
        self.owner.friends.add(self.requester.user)
        self.assertTrue(
            is_access_allowed(self.owner, self.requester, self.obj),
            msg=('Access should be allowed, if the requester is at least a'
                 ' friend.'))

        # Level 4
        setting.level = self.level_4
        setting.save()
        self.assertFalse(
            is_access_allowed(self.owner, self.requester, self.obj),
            msg=('Access should only be allowed, if the requester is also the'
                 ' owner.'))
        self.assertTrue(
            is_access_allowed(self.owner, self.owner, self.obj),
            msg=('Access should only be allowed, if the requester is also the'
                 ' owner.'))

    def test_tag_with_field_name(self):
        # No level
        self.assertTrue(
            is_access_allowed(self.owner, self.requester, self.obj, 'name'),
            msg=('Access should be allowed, if it\'s an object without privacy'
                 ' setting.'))

        # Level 1
        setting = PrivacySettingFactory(
            level=self.level_1, content_object=self.obj, field_name='name')
        self.assertTrue(
            is_access_allowed(self.owner, self.requester, self.obj, 'name'),
            msg=('Access should be allowed, if it\'s a public object.'))

        # Level 2
        setting.level = self.level_2
        setting.save()
        self.assertFalse(
            is_access_allowed(self.owner, self.requester, self.obj, 'name'),
            msg=('Access should not be allowed, if the requester is not at'
                 ' least a friend of a friend.'))
        friend = DummyProfileModelFactory()
        self.owner.friends.add(friend.user)
        self.requester.friends.add(friend.user)
        self.assertTrue(
            is_access_allowed(self.owner, self.requester, self.obj, 'name'),
            msg=('Access should be allowed, if the requester is at least a'
                 ' friend of a friend.'))



"""Template tags of the ``privacy`` app."""
from django.conf import settings
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ObjectDoesNotExist
from django.template import Library

from django_libs.loaders import load_member_from_setting

from ..models import PrivacyLevel, PrivacySetting

register = Library()


def get_privacy_dict(obj, field_name):
    """
    Function, which returns a dictionary of relevant data of a privacy level
    (field).

    """
    kwargs = {
        'content_type': ContentType.objects.get_for_model(obj),
        'object_id': obj.pk,
    }
    if field_name:
        kwargs.update({'field_name': field_name})
    else:
        kwargs.update({'field_name': ''})
    privacy_levels = PrivacyLevel.objects.all()
    try:
        selected_setting = PrivacySetting.objects.get(**kwargs)
        selected_level = selected_setting.level.clearance_level
    except PrivacySetting.DoesNotExist:
        selected_setting = None
        selected_level = getattr(
            settings, 'PRIVACY_DEFAULT_CLEARANCE_LEVEL', 1)
    return {
        'field_name': field_name,
        'privacy_levels': privacy_levels,
        'selected_level': selected_level,
        'selected_setting': selected_setting,
    }


@register.assignment_tag
def get_privacy_setting(obj, field_name=None):
    privacy_dict = get_privacy_dict(obj, field_name)
    return privacy_dict['selected_setting']


@register.inclusion_tag('privacy/partials/field.html')
def render_privacy_level_field(obj, field_name=None):
    return get_privacy_dict(obj, field_name)


def is_access_allowed(owner, requester, obj, field_name=None):
    get_clearance_level = load_member_from_setting(
        'PRIVACY_CLEARANCE_LEVEL_FUNCTION')
    clearance_level = get_clearance_level(owner, requester)
    kwargs = {
        'content_type': ContentType.objects.get_for_model(obj),
        'object_id': obj.pk,
    }
    if field_name:
        kwargs.update({'field_name': field_name})
    else:
        kwargs.update({'field_name': ''})
    try:
        privacy_setting = PrivacySetting.objects.get(**kwargs)
    except ObjectDoesNotExist:
        return True  # No settings given? Allow for everyone
    if privacy_setting.level.clearance_level <= clearance_level:
        # Level has to be at least as high as the setting's label of the
        # object/field
        return True
    return False

register.assignment_tag(is_access_allowed)



"""Privacy settings for the ``test_app`` app."""


def get_clearance_level(owner, requester):
    """
    4 = self
    3 = friend
    2 = friend of friend
    1 = public

    """
    if owner == requester:
        return 4
    if requester.user in owner.friends.all():
        return 3

    # Get friends of friends
    owner_friends_pks = owner.friends.values_list('pk')
    requester_friends_pks = requester.friends.values_list('pk')
    if set(owner_friends_pks) & set(requester_friends_pks):
        return 2
    return 1



"""Factories for the ``test_app`` app."""
import factory

from django_libs.tests.factories import UserFactory

from .models import DummyModel, DummyProfileModel


class DummyModelFactory(factory.DjangoModelFactory):
    FACTORY_FOR = DummyModel

    name = factory.Sequence(lambda x: 'A name {0}'.format(x))


class DummyProfileModelFactory(factory.DjangoModelFactory):
    FACTORY_FOR = DummyProfileModel

    user = factory.SubFactory(UserFactory)



"""Factories for the ``privacy`` app."""
import factory

from django_libs.tests.factories import HvadFactoryMixin

from ..models import PrivacyLevel, PrivacySetting
from test_app.factories import DummyModelFactory


class PrivacyLevelFactory(HvadFactoryMixin, factory.DjangoModelFactory):
    FACTORY_FOR = PrivacyLevel

    language_code = 'en'
    name = factory.Sequence(lambda x: 'A name {0}'.format(x))
    clearance_level = factory.Sequence(lambda x: x)


class PrivacySettingFactory(factory.DjangoModelFactory):
    FACTORY_FOR = PrivacySetting

    level = factory.SubFactory(PrivacyLevelFactory)
    content_object = factory.SubFactory(DummyModelFactory)



"""Tests for the forms of the ``privacy`` app."""
from django.test import TestCase

from .factories import PrivacyLevelFactory, PrivacySettingFactory
from ..models import PrivacySetting
from test_app.factories import DummyProfileModelFactory
from test_app.forms import DummyProfileModelForm


class DummyProfileModelFormTestCase(TestCase):
    """Tests for the ``DummyProfileModelForm`` form."""
    longMessage = True

    def setUp(self):
        self.obj = DummyProfileModelFactory()
        self.level_1 = PrivacyLevelFactory(clearance_level=1)
        self.level_2 = PrivacyLevelFactory(clearance_level=2)
        self.level_3 = PrivacyLevelFactory(clearance_level=3)
        self.level_4 = PrivacyLevelFactory(clearance_level=4)
        self.field_setting = PrivacySettingFactory(
            content_object=self.obj, level=self.level_1, field_name='name')

    def test_tag(self):
        data = {
            'user': self.obj.user.pk,
            'privacy_for_instance': 3,
        }
        form = DummyProfileModelForm(instance=self.obj, data=data)
        self.assertTrue(form.is_valid(), msg=('Form should be valid.'))
        form.save()
        self.setting = PrivacySetting.objects.get(
            object_id=self.obj.pk, field_name='')
        self.assertEqual(self.setting.level, self.level_3, msg=(
            'The form instance\'s privacy setting should be set to 3.'))
        data.update({'privacy_name': 4})
        form = DummyProfileModelForm(instance=self.obj, data=data)
        self.assertTrue(form.is_valid(), msg=('Form should be valid.'))
        form.save()
        self.assertEqual(
            PrivacySetting.objects.get(pk=self.field_setting.pk).level,
            self.level_4,
            msg=('The name field\'s privacy setting should be set to 3.'))
